(**************************************************************************)
(*                                                                        *)
(*  Setup for lightweight web server                                      *)
(*  Copyright (C) 2024   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE DocRoot;

        (****************************************************************)
        (*                                                              *)
        (*                    PM Setup for WebServe                     *)
        (*              First page of the Domain notebook               *)
        (*                                                              *)
        (*        Started:        14 December 2024                      *)
        (*        Last edited:    18 December 2024                      *)
        (*        Status:         OK                                    *)
        (*                                                              *)
        (****************************************************************)


FROM SYSTEM IMPORT ADDRESS, INT16, CAST, ADR;

IMPORT OS2, OS2RTL, SYSTEM;

IMPORT DID, Strings, CommonSettings, INIData, RINIData, OneLine;

FROM Languages IMPORT
    (* type *)  LangHandle,
    (* proc *)  StrToBuffer, StrToBufferN;

FROM RINIData IMPORT
    (* proc *)  RemoteOperation, INIPut, INIPutString,
                OpenINIFile, CloseINIFile, INIDeleteKey,
                INIFetch, INIGetCard, INIGetTwoShort, INIGetString,
                MakeDirectory;

FROM Names IMPORT
    (* type *)  DomainName, HostNameIndex, HostName, FilenameIndex, FilenameString;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM Remote IMPORT
    (* proc *)  OurDirectory;

FROM MiscFuncs IMPORT
    (* type *)  CharArrayPointer,
    (* proc *)  EVAL;

(**************************************************************************)

CONST
    Nul = CHR(0);
    LangStringSize = 32;

TYPE
    ServiceType = (HTTP, HTTPS);
    CardArray = ARRAY ServiceType OF CARDINAL;

VAR
    INIname, SetupINIname: ARRAY [0..31] OF CHAR;
    domname: DomainName;

    ChangeInProgress: BOOLEAN;
    OurPageID: CARDINAL;
    OurLang: LangHandle;

    (* Handle to the window that belongs to this page.  We can save     *)
    (* this as a static value because there is never more than one      *)
    (* instance of this page.                                           *)

    pagehandle, hwndParent: OS2.HWND;
    notebookhandle: OS2.HWND;
    PageFont: CommonSettings.FontName;

    (* Original values of some ini variables. *)

    OldHTMLRoot, OldCGIDir: FilenameString;

(************************************************************************)
(*                    OPERATIONS ON DIALOGUE LABELS                     *)
(************************************************************************)

PROCEDURE SetLanguage (hwnd: OS2.HWND;  lang: LangHandle);

    (* Relabels this page in the new language. *)

    VAR stringval: ARRAY [0..511] OF CHAR;

    BEGIN
        OurLang := lang;
        StrToBuffer (lang, "HTML.tab", stringval);
        OS2.WinSendMsg (notebookhandle, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,OurPageID), ADR(stringval));
        StrToBuffer (lang, "HTML.RootLabel", stringval);
        OS2.WinSetDlgItemText (hwnd, DID.RootLabel, stringval);
        StrToBuffer (lang, "HTML.CGIDirLabel", stringval);
        OS2.WinSetDlgItemText (hwnd, DID.CGIDirLabel, stringval);
        StrToBuffer (lang, "HTML.HostlistLabel", stringval);
        OS2.WinSetDlgItemText (hwnd, DID.HostlistLabel, stringval);
        StrToBuffer (lang, "Buttons.Add", stringval);
        OS2.WinSetDlgItemText (hwnd, DID.Addhost, stringval);
        StrToBuffer (lang, "Buttons.Promote", stringval);
        OS2.WinSetDlgItemText (hwnd, DID.Promotehost, stringval);
        StrToBuffer (lang, "Buttons.Delete", stringval);
        OS2.WinSetDlgItemText (hwnd, DID.Deletehost, stringval);
    END SetLanguage;

(************************************************************************)

PROCEDURE SetPageFonts;

    (* Changes the font of the notebook pages to the font recorded in   *)
    (* the INI file as belonging to this dialogue.                      *)

    VAR NewFontName: CommonSettings.FontName;

    BEGIN
        CommonSettings.CurrentFont (CommonSettings.DomainNotebook, NewFontName);
        IF NOT Strings.Equal (NewFontName, PageFont) THEN
            PageFont := NewFontName;
        END (*IF*);
    END SetPageFonts;

(************************************************************************)
(*                     LOADING DATA FROM THE INI FILE                   *)
(************************************************************************)

PROCEDURE LoadValues (hwnd: OS2.HWND;  domain: ARRAY OF CHAR);

    (* Fills the dialogue elements on page 1 with data from the INI file,       *)
    (* or loads default values if they're not in the INI file.                  *)

    VAR str: FilenameString;
        state: RINIData.StringReadState;
        name: HostName;
        namesset: BOOLEAN;

    BEGIN
        EVAL (OpenINIFile (INIname));

        (* HTML root. *)

        IF RINIData.INIGetString (domain, 'HTMLRoot', OldHTMLRoot) THEN
            str := OldHTMLRoot;
        ELSE
            OldHTMLRoot := "xxx";
            str := ".";
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.HTMLRoot, str);

        (* cgi-bin directory. *)

        IF RINIData.INIGetString (domain, 'CGIDir', OldCGIDir) THEN
            str := OldCGIDir;
        ELSE
            OldCGIDir := "xxx";
            str := "cgi-bin";
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.CGIDir, str);

        (* The list of alternative names for this domain. *)

        IF RINIData.INIFetch (domain, "namesset", namesset) THEN
            RINIData.GetStringList (domain, "names", state);
            REPEAT
                RINIData.NextString (state, name);
                IF name[0] <> Nul THEN
                    (* Add name to the listbox. *)

                    OS2.WinSendDlgItemMsg (hwnd, DID.Hostlist, OS2.LM_INSERTITEM,
                         OS2.MPFROMSHORT(OS2.LIT_END), SYSTEM.ADR(name));

                END (*IF*);
            UNTIL name[0] = Nul;
            RINIData.CloseStringList (state);
        ELSE
            (* First time, so add a wildcard entry. *)

            name := "*.";
            Strings.Append (domname, name);
            OS2.WinSendDlgItemMsg (hwnd, DID.Hostlist, OS2.LM_INSERTITEM,
                         OS2.MPFROMSHORT(OS2.LIT_END), SYSTEM.ADR(name));
        END (*IF*);


        CloseINIFile;

    END LoadValues;

(************************************************************************)
(*                      STORING DATA TO THE INI FILE                    *)
(************************************************************************)

(*
PROCEDURE QueryButton (hwnd: OS2.HWND;  B: CARDINAL): CARDINAL;

    BEGIN
        RETURN OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, B,
                                              OS2.BM_QUERYCHECK, NIL, NIL));
    END QueryButton;
*)

(**************************************************************************)

PROCEDURE StoreList (hwnd: OS2.HWND);

    (* Stores a HostList into the INI file.  The hwnd parameter is the  *)
    (* handle of the listbox.  We assume that the INI file is already   *)
    (* open.                                                            *)

    CONST NameLength = MAX(HostNameIndex)+1;

    VAR bufptr: CharArrayPointer;
        BufferSize: CARDINAL;
        j, k, count, index: CARDINAL;
        name: HostName;
        namesset: BOOLEAN;

    BEGIN
        (*EVAL (OpenINIFile (INIname));*)
        (* Work out how much buffer space we need. *)

        BufferSize := 0;
        count := OS2.ULONGFROMMR(OS2.WinSendMsg (hwnd, OS2.LM_QUERYITEMCOUNT, NIL, NIL));
        IF count > 0 THEN
            FOR index := 0 TO count-1 DO
                INC (BufferSize,
                     OS2.ULONGFROMMR(OS2.WinSendMsg (hwnd, OS2.LM_QUERYITEMTEXTLENGTH,
                                   OS2.MPFROMUSHORT(index), NIL)) + 1);
            END (*FOR*);
        END (*IF*);

        (* Create the string buffer. *)

        IF BufferSize = 0 THEN
            bufptr := NIL;
        ELSE
            INC (BufferSize);
            ALLOCATE (bufptr, BufferSize);
        END (*IF*);

        (* Store all the strings into the buffer. *)

        IF count > 0 THEN
            j := 0;
            FOR index := 0 TO count-1 DO
                OS2.WinSendMsg (hwnd, OS2.LM_QUERYITEMTEXT,
                                OS2.MPFROM2USHORT(index, NameLength), SYSTEM.ADR(name));
                k := 0;
                REPEAT
                    bufptr^[j] := name[k];
                    INC (k);  INC (j);
                UNTIL (name[k] = Nul) OR (k = NameLength);
                bufptr^[j] := Nul;
                INC (j);
            END (*FOR*);

            bufptr^[j] := Nul;

        END (*IF*);

        (* Write the buffer to the INI file. *)

        IF BufferSize = 0 THEN
            RINIData.INIPutBinary (domname, "names", j, 0);
        ELSE
            RINIData.INIPutBinary (domname, "names", bufptr^, BufferSize);
        END (*IF*);
        namesset := TRUE;
        RINIData.INIPut (domname, "namesset", namesset);

        (* Deallocate the buffer space. *)

        IF BufferSize > 0 THEN
            DEALLOCATE (bufptr, BufferSize);
        END (*IF*);

    END StoreList;

(************************************************************************)

PROCEDURE StoreData (hwnd: OS2.HWND;  domain: ARRAY OF CHAR);

    (* Stores the values on this page back into the INI file.  *)

    VAR val: FilenameString;

    BEGIN
        EVAL (RINIData.OpenINIFile (INIname));

        (* HTML root. *)

        OS2.WinQueryDlgItemText (hwnd, DID.HTMLRoot, MAX(FilenameIndex)+1, val);
        IF val <> OldHTMLRoot THEN
            RINIData.INIPutString (domain, "HTMLRoot", val);
        END (*IF*);

        (* cgi-bin directory. *)

        OS2.WinQueryDlgItemText (hwnd, DID.CGIDir, MAX(FilenameIndex)+1, val);
        IF val <> OldCGIDir THEN
            RINIData.INIPutString (domain, "CGIDir", val);
        END (*IF*);

        (* Alternative names for the domain. *)

        StoreList (OS2.WinWindowFromID (hwnd, DID.Hostlist));

        RINIData.CloseINIFile;

    END StoreData;

(************************************************************************)
(*                WINDOW PROCEDURE FOR SUBCLASSED CASE                  *)
(************************************************************************)

PROCEDURE ["SysCall"] SubWindowProc (hwnd     : OS2.HWND;
                                     msg      : OS2.ULONG;
                                     mp1, mp2 : OS2.MPARAM): OS2.MRESULT;

    (* Window procedure to intercept some of the things that happen in  *)
    (* the notebook subwindow.  We want this here mainly so that we can *)
    (* detect a new font dropped on the notebook tabs.  If the message  *)
    (* is something we don't want to deal with here, we pass it         *)
    (* to the parent window procedure.                                  *)

    VAR OldWndProc: OS2.PFNWP;
        owner: OS2.HWND;
        length, AttrFound: CARDINAL;
        NewFontName: CommonSettings.FontName;

    BEGIN
        OldWndProc := SYSTEM.CAST (OS2.PFNWP, OS2.WinQueryWindowPtr (hwnd, OS2.QWL_USER));
        owner := OS2.WinQueryWindow(hwnd,OS2.QW_OWNER);

        (* Because of the interaction between subclassing and DragText, *)
        (* some messages will go lost if we use the obvious strategy of *)
        (* sending them through to OldWndProc.  To get around this, we  *)
        (* have to send those messages directly to the target window.   *)

        IF (msg = OS2.WM_BUTTON2DOWN) OR (msg = OS2.DM_DRAGOVER)
                   OR (msg = OS2.DM_DRAGLEAVE) OR (msg = OS2.DM_DROP) THEN

            RETURN OS2.WinSendMsg (owner, msg, mp1, mp2);

        (* Check for font or colour change. *)

        ELSIF msg = OS2.WM_PRESPARAMCHANGED THEN

            IF ChangeInProgress THEN
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            ELSE
                ChangeInProgress := TRUE;
                length := OS2.WinQueryPresParam (hwnd, OS2.PP_FONTNAMESIZE, 0,
                                             AttrFound, CommonSettings.FontNameSize, NewFontName,
                                              0(*OS2.QPF_NOINHERIT*));
                IF length < CommonSettings.FontNameSize THEN
                    NewFontName[length] := Nul;
                END (*IF*);
                ChangeInProgress := FALSE;
                RETURN NIL;
            END (*IF*);

        END (*IF*);

        RETURN OldWndProc (hwnd, msg, mp1, mp2);

    END SubWindowProc;

(************************************************************************)
(*                   WINDOW PROCEDURE FOR MAIN DIALOGUE                 *)
(************************************************************************)

PROCEDURE ["SysCall"] DialogueProc (hwnd: OS2.HWND;  msg: OS2.ULONG;
                                      mp1, mp2: OS2.MPARAM): OS2.MRESULT;

    VAR listwindow: OS2.HWND;
        ButtonID, NotificationCode: CARDINAL;
        index: INTEGER;
        name: ARRAY HostNameIndex OF CHAR;
        message: ARRAY [0..127] OF CHAR;

    BEGIN
        index := OS2.LONGFROMMR(
                   OS2.WinSendDlgItemMsg (hwnd, DID.Hostlist, OS2.LM_QUERYSELECTION, NIL, NIL));
        CASE msg OF
           |  OS2.WM_INITDLG:
                   OS2.WinSetWindowPtr (hwnd, OS2.QWL_USER,
                               SYSTEM.CAST(SYSTEM.ADDRESS,
                                    OS2.WinSubclassWindow (hwnd,
                                                           SubWindowProc)));
                   OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.Promotehost), FALSE);
                   OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.Deletehost), FALSE);
                   LoadValues (hwnd, domname);

           |  CommonSettings.FONTCHANGED:

                   IF ChangeInProgress THEN
                       RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                   ELSE
                       ChangeInProgress := TRUE;
                       SetPageFonts;
                       ChangeInProgress := FALSE;
                       RETURN NIL;
                   END (*IF*);

           |  OS2.WM_COMMAND:

                   listwindow := OS2.WinWindowFromID(hwnd,DID.Hostlist);
                   ButtonID := OS2.SHORT1FROMMP(mp1);

                   IF ButtonID = DID.Addhost THEN
                          name := "";
                          StrToBuffer (OurLang, "Local.EnterName", message);
                          OneLine.Edit (hwnd, SetupINIname, message, name);
                          IF name[0] <> Nul THEN
                              IF index = OS2.LIT_NONE THEN
                                  index := 0;
                              ELSE
                                  INC(index);
                              END (*IF*);
                              OS2.WinSendDlgItemMsg (hwnd, DID.Hostlist, OS2.LM_INSERTITEM,
                                     OS2.MPFROMSHORT(index), SYSTEM.ADR(name));
                              OS2.WinSendDlgItemMsg (hwnd, DID.Hostlist, OS2.LM_SELECTITEM,
                                     OS2.MPFROMSHORT(index), OS2.MPFROMSHORT(ORD(TRUE)));
                          END (*IF*);

                   ELSIF ButtonID = DID.Promotehost THEN

                          OS2.WinSendMsg (listwindow, OS2.LM_QUERYITEMTEXT,
                                          OS2.MPFROM2USHORT(index, MAX(HostNameIndex)+1), SYSTEM.ADR(name));
                          OS2.WinSendMsg (listwindow, OS2.LM_DELETEITEM,
                                                 OS2.MPFROMSHORT(index), NIL);
                          DEC (index);
                          OS2.WinSendMsg (listwindow, OS2.LM_INSERTITEM,
                                 OS2.MPFROMSHORT(index), SYSTEM.ADR(name));
                          OS2.WinSendMsg (listwindow, OS2.LM_SELECTITEM,
                                 OS2.MPFROMSHORT(index), OS2.MPFROMSHORT(ORD(TRUE)));

                   ELSIF ButtonID = DID.Deletehost THEN

                          OS2.WinSendDlgItemMsg (hwnd, DID.Hostlist, OS2.LM_DELETEITEM,
                                                 OS2.MPFROMSHORT(index), NIL);
                          OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.Promotehost), FALSE);
                          OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.Deletehost), FALSE);

                   END (*IF*);
                   RETURN NIL;

           |  OS2.WM_CONTROL:

                   NotificationCode := OS2.ULONGFROMMP(mp1);
                   ButtonID := NotificationCode MOD 65536;
                   NotificationCode := NotificationCode DIV 65536;

                   IF ButtonID = DID.Hostlist THEN
                       IF NotificationCode = OS2.LN_SELECT THEN

                           (* For some reason the more obvious code doesn't work below, so     *)
                           (* we have to use an if/then/else construct.                        *)

                           IF index > 0 THEN
                               OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.Promotehost), TRUE);
                           ELSE
                               OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.Promotehost), FALSE);
                           END (*IF*);
                           OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.Deletehost), TRUE);
                           RETURN NIL;
                       ELSIF NotificationCode = OS2.LN_ENTER THEN
                           (* Ignore this one for now, maybe eliminate later. *)
                           RETURN NIL;
                       ELSE
                           RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                       END (*IF*);
                   ELSE
                       RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                   END (*IF*);

           |  OS2.WM_CLOSE:
                   CommonSettings.EnableFontChanges(FALSE);
                   RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);

        ELSE    (* default *)
           RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
        END (*CASE*);
        RETURN NIL;

    END DialogueProc;

(**************************************************************************)

PROCEDURE CreatePage (notebook: OS2.HWND;  VAR (*IN*) domainname: ARRAY OF CHAR;
                        VAR (*OUT*) PageID: CARDINAL): OS2.HWND;

    (* Creates page 1 and adds it to the notebook. *)

    VAR Label: ARRAY [0..31] OF CHAR;

    BEGIN
        Strings.Assign (domainname, domname);
        CommonSettings.GetINIFilename (INIname);
        CommonSettings.GetSetupINIname (SetupINIname);
        notebookhandle := notebook;
        hwndParent := OS2.WinQueryWindow (notebook, OS2.QW_PARENT);
        pagehandle := OS2.WinLoadDlg(notebook, notebook,
                       DialogueProc,    (* dialogue procedure *)
                       0,                   (* use resources in EXE *)
                       DID.DocRoot,          (* dialogue ID *)
                       NIL);                 (* creation parameters *)
        PageID := OS2.ULONGFROMMR (OS2.WinSendMsg (notebook, OS2.BKM_INSERTPAGE,
                         NIL,
                         OS2.MPFROM2SHORT (OS2.BKA_MAJOR+OS2.BKA_AUTOPAGESIZE, OS2.BKA_LAST)));
        OurPageID := PageID;
        Label := "Page 1";
        OS2.WinSendMsg (notebook, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,PageID), ADR(Label));
        OS2.WinSendMsg (notebook, OS2.BKM_SETPAGEWINDOWHWND,
                        CAST(ADDRESS,PageID), CAST(ADDRESS,pagehandle));
        RETURN pagehandle;
    END CreatePage;

(**************************************************************************)

PROCEDURE SetFont (VAR (*IN*) name: CommonSettings.FontName);

    (* Sets the font of the text on this page. *)

    CONST bufsize = CommonSettings.FontNameSize;

    BEGIN
        OS2.WinSetPresParam (pagehandle, OS2.PP_FONTNAMESIZE, bufsize, name);
    END SetFont;

(************************************************************************)

PROCEDURE Close (notebook, hwnd: OS2.HWND);

    (* Shuts down this window and removes it from the notebook. *)

    BEGIN
        StoreData (hwnd, domname);
        OS2.WinSendMsg (notebook, OS2.BKM_DELETEPAGE,
                        CAST(ADDRESS, OurPageID),
                        OS2.MPFROMLONG (OS2.BKA_SINGLE));
        OS2.WinSendMsg (hwnd, OS2.WM_CLOSE, NIL, NIL);
    END Close;

(**************************************************************************)

BEGIN
    INIname := "garbage.INI";
    ChangeInProgress := FALSE;
    pagehandle := OS2.NULLHANDLE;
    hwndParent := OS2.NULLHANDLE;
END DocRoot.

