(**************************************************************************)
(*                                                                        *)
(*  Socket I/O with TLS support                                           *)
(*  Copyright (C) 2025   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE NetStream;

        (********************************************************)
        (*                                                      *)
        (*          Handler for I/O through either a TLS        *)
        (*          channel or directly through a socket        *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            2 April 2018                    *)
        (*  Last edited:        1 March 2025                    *)
        (*  Status:             Working if you don't use TLS    *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM SYSTEM IMPORT ADR, CARD8, LOC;

FROM TLS IMPORT
    (* type *)  TLSsession,
    (* proc *)  OpenTLSsession, CloseTLSsession, TLSsend,
                TLSrecv, TLSrecvLine;

IMPORT SockStream;

FROM SockStream IMPORT
    (* type *)  SSstate,
    (* proc *)  OpenSockStream, PutToSocket;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransactionL;

FROM Inet2Misc IMPORT
    (* proc *)  WaitForSocketOut, Synch;

FROM Sockets IMPORT
    (* type *)  Socket,
    (* proc *)  recv, send;

FROM Watchdog IMPORT
    (* type *)  WatchdogID,
    (* proc *)  KickWatchdog;

FROM LowLevel IMPORT
    (* proc *)  Copy, EVAL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    BufferSize = 512;
    Nul = CHR(0);  CR = CHR(13);  LF = CHR(10);

TYPE
    NStream = POINTER TO RECORD
                             TLS, IsServer: BOOLEAN;
                             socket: Socket;
                             logID: TransactionLogID;
                             TLShandle: TLSsession;
                             SStream: SSstate;
                         END (*RECORD*);

(************************************************************************)
(*                              OPEN/CLOSE                              *)
(************************************************************************)

PROCEDURE OpenNetStream (S: Socket;  IsServer: BOOLEAN;
                        VAR (*IN*) serverdomain: ARRAY OF CHAR;
                        logID: TransactionLogID;  TLS: BOOLEAN): NStream;

    (* Opens a new NStream object that handles socket data. *)

    (* If as_server is FALSE, the serverdomain parameter specifies the  *)
    (* domain the client is trying to contact. Otherwise, it specifies  *)
    (* an initial value for the domain the server is hosting, but in    *)
    (* that case it will probably be overridden by the domain           *)
    (* specified in a ClientHello message.  If TLS is FALSE then the    *)
    (* serverdomain parameter is ignored.                               *)

    VAR result: NStream;

    BEGIN
        NEW (result);
        result^.TLS := TLS;
        result^.IsServer := IsServer;
        result^.socket := S;
        result^.logID := logID;
        IF result^.TLS THEN
            IF OpenTLSsession(S, IsServer, serverdomain, logID,
                                            result^.TLShandle) THEN
                result^.SStream := NIL;
            ELSE
                LogTransactionL (logID, "TLS handshaking failed");
                result^.TLS := FALSE;
            END (*IF*);
        END (*IF*);
        IF NOT result^.TLS THEN
            result^.TLShandle := NIL;
            result^.SStream := OpenSockStream (S);
        END (*IF*);
        RETURN result;
    END OpenNetStream;

(************************************************************************)

PROCEDURE CloseNetStream (VAR (*INOUT*) stream: NStream);

    (* Deletes the stream. *)

    BEGIN
        IF stream <> NIL THEN
            IF stream^.TLS THEN
                CloseTLSsession (stream^.TLShandle);
            ELSE
                SockStream.CloseSockStream (stream^.SStream);
            END (*IF*);
            (*DISPOSE (stream);*)
        END (*IF*);
    END CloseNetStream;

(************************************************************************)

PROCEDURE StartTLS (stream: NStream;  VAR (*IN*) serverdomain: ARRAY OF CHAR);

    (* Switches from plain mode to TLS mode.  Assumption: the stream    *)
    (* has already been opened with TLS = FALSE.  There is no way of    *)
    (* switching out of this mode again.                                *)

    BEGIN
        IF (stream = NIL) OR stream^.TLS THEN
            (* Operation fails - illegal call. *)
        ELSE
            WITH stream^ DO
                TLS := OpenTLSsession(socket, IsServer, serverdomain,
                                                       logID, TLShandle);
                IF NOT TLS THEN
                    LogTransactionL (logID, "TLS handshaking failed");
                END (*IF*);
            END (*WITH*);
        END (*IF*);
    END StartTLS;

(************************************************************************)
(*                            SYNCHRONISATION                           *)
(************************************************************************)

(*
PROCEDURE DoSynch (stream: NStream);

    (* Forces any pending output to be sent. *)

    BEGIN
        Synch (stream^.sock);
    END DoSynch;
*)

(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetLine (stream: NStream;  VAR (*OUT*) line: ARRAY OF CHAR): BOOLEAN;

    (* Gets one line from the specified stream.  Returns a FALSE        *)
    (* result, with no data, at end of stream.                          *)

    (* To note: we also have no data in the case of a blank line, but   *)
    (* in that case we return a TRUE result.                            *)

    VAR actual: CARDINAL;  success, ConnectionLost: BOOLEAN;

    BEGIN
        IF stream^.TLS THEN
            success := TLSrecvLine (stream^.TLShandle, line,
                        HIGH(line)+1, actual);
        ELSE
            success := SockStream.GetLine (stream^.SStream, line);
        END (*IF*);
        (*
        IF NOT success THEN
            IF stream^.IsServer THEN
                LogTransactionL (stream^.logID, "Connection closed by client");
            ELSE
                LogTransactionL (stream^.logID, "Connection closed by server");
            END (*IF*);
            CloseNetStream (stream);
        END (*IF*);
        *)
        RETURN success;
    END GetLine;

(************************************************************************)

PROCEDURE GetBytes (stream: NStream;  VAR (*OUT*) result: ARRAY OF CHAR;
                        wanted: CARDINAL;  VAR (*OUT*) actual: CARDINAL);

    (* Gets at most wanted chars into result,  Returns actual as the    *)
    (* number of characters found.                                      *)

    VAR ConnectionLost: BOOLEAN;

    BEGIN
        IF stream^.TLS THEN
            TLSrecv (stream^.TLShandle, result,
                            wanted, actual, ConnectionLost);
        ELSE
            actual := SockStream.PutToSocket (stream^.SStream, result, wanted);
        END (*IF*);
    END GetBytes;

(************************************************************************)
(*                                  OUTPUT                              *)
(************************************************************************)

PROCEDURE PutBytes (stream: NStream;  VAR (*IN*) message: ARRAY OF LOC;
                                                length: CARDINAL): CARDINAL;

    (* Sends a message of "length" bytes.  The returned value is the number *)
    (* of bytes sent, or MAX(CARDINAL) if there was an error.               *)

    BEGIN
        IF stream^.TLS THEN
            TLSsend (stream^.TLShandle, message, length);
            RETURN length;
        ELSE
            RETURN PutToSocket (stream^.SStream, message, length);
        END (*IF*);
    END PutBytes;

(************************************************************************)

PROCEDURE PutLine (stream: NStream;  VAR (*IN*) message: ARRAY OF LOC;
                                                length: CARDINAL): CARDINAL;

    (* Sends a message of "length" bytes, followed by CRLF.  The        *)
    (* returned value is the number of bytes sent, or MAX(CARDINAL) if  *)
    (* there was an error.                                              *)

    VAR buffer: ARRAY [0..1023] OF CARD8;

    BEGIN
        Copy (ADR(message), ADR(buffer), length);
        buffer[length] := ORD(CR);  INC (length);
        buffer[length] := ORD(LF);  INC (length);
        RETURN PutBytes (stream, buffer, length);
    END PutLine;

(************************************************************************)

PROCEDURE PutEOL (stream: NStream);

    (* Sends CRLF.  *)

    TYPE TwoByte = ARRAY [0..1] OF CHAR;
    CONST EOL = TwoByte {CR, LF};
    VAR CRLF: TwoByte;

    BEGIN
        CRLF := EOL;
        EVAL (PutBytes (stream, CRLF, 2));
    END PutEOL;

(************************************************************************)

END NetStream.

