(**************************************************************************)
(*                                                                        *)
(*  Transport Layer Security                                              *)
(*  Copyright (C) 2025   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TLSsessID;

        (********************************************************)
        (*                                                      *)
        (*                Transport Layer Security              *)
        (*                Cacheing of session IDs               *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            25 January 2025                 *)
        (*  Last edited:        10 February 2025                *)
        (*  Status:             Done, untested                  *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(*                                                                      *)
(* If the server accepts a session ID offered by the client, it must be *)
(* prepared to proceed imnmediately to a ServerHello followed by an     *)
(* optional ChangeCipherSpec and a Finished message.  (The client will  *)
(* respond with its own ChangeCipherSpec and Finished.)  What does this *)
(* imply about how much information must be stored with a saved ID?     *)
(* To do a ChangeCipherSpec, it turns out to be sufficent to save the   *)
(* ciphersuite and the master secret.  When those are restored, we can  *)
(* generate a new key block before changing the cipher spec.            *)
(*                                                                      *)
(* I deduced these details from RFC 5077.                               *)
(*                                                                      *)
(************************************************************************)

FROM SYSTEM IMPORT CARD8, ADR;

IMPORT Strings, INIData;

FROM TLSCipherSuites IMPORT
    (* type *)  CipherSuite;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  DiscardBS;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, Obtain, Release;

FROM LowLevel IMPORT
    (* proc *)  Copy, IXORB;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST mslength = 48;  maxcount = 50;  Nul = CHR(0);

TYPE
    SavedSess = POINTER TO
                    RECORD
                        prev, next: SavedSess;
                        sessID: sessIDtype;
                        suite: CipherSuite;
                        master: ARRAY [0..mslength-1] OF CARD8;
                    END (*RECORD*);

VAR NextID: sessIDtype;
    NextIDLock: Lock;
    sesslist:   RECORD
                    access: Lock;
                    count: CARDINAL;
                    head, tail: SavedSess;
                END (*RECORD*);

    (* Our INI file name. *)

    ININame: ARRAY [0..31] OF CHAR;

(************************************************************************)
(*                 CHECKSUM (ONLY NEEDED FOR DEBUGGING)                 *)
(************************************************************************)

PROCEDURE BlockChecksum (N: CARDINAL;  VAR (*IN*) A: ARRAY OF CARD8): CARD8;

    (* Calculates a checksum for an array of N bytes. *)

    VAR result: CARD8;  j: CARDINAL;

    BEGIN
        result := 0;
        FOR j := 0 TO N-1 DO
            result := IXORB (result, A[j]);
        END (*FOR*);
        RETURN result;
    END BlockChecksum;

(************************************************************************)
(*                   MISC OPERATIONS ON SESSION ID                      *)
(************************************************************************)

PROCEDURE ResetNextID;

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO lastAN DO
            NextID[j] := '0';
        END (*FOR*);
        NextID[lastAN+1] := Nul;
    END ResetNextID;

(************************************************************************)

PROCEDURE FindID (VAR (*IN*) ID: ARRAY OF CHAR): SavedSess;

    (* Returns a pointer to the saved record for session ID.    *)
    (* Returns NIL if no such record is found.  The caller      *)
    (* must have exclusive access to the list.                  *)

    VAR p: SavedSess;

    BEGIN
        p := sesslist.head;
        WHILE (p <> NIL) AND NOT Strings.Equal(p^.sessID, ID) DO
            p := p^.next;
        END (*WHILE*);
        RETURN p;
    END FindID;

(************************************************************************)

PROCEDURE LastSavedID (VAR (*OUT*) sessID: sessIDtype);

    (* Returns the ID of the last saved session, or "" if none found. *)

    VAR p: SavedSess;

    BEGIN
        Obtain (NextIDLock);
        p := sesslist.head;
        IF p = NIL THEN
            sessID := "";
        ELSE
            Strings.Assign (p^.sessID, sessID);
        END (*KF*);
        Release (NextIDLock);
    END LastSavedID;

(************************************************************************)

PROCEDURE IncrNextID (k: CARDINAL);

    (* Increments NextID[0..k].  We assume that the caller has          *)
    (* exclusive access to NextID.                                      *)

    (* The standard allows for 32 bytes, i.e. 256 bits, of session ID,  *)
    (* but I am never going to save as many as 2^256 sessions.  So I'm  *)
    (* only using alphanumeric characters, and I'm only using locations *)
    (* [0..lastAN] of the ID, plus one trailing nul.                    *)

    BEGIN
        IF NextID[k] >= 'z' THEN
            NextID[k] := '0';
            IF k = 0 THEN
                ResetNextID;
            ELSE
                IncrNextID (k-1);
            END (*IF*);
        ELSIF NextID[k] = 'Z' THEN
            NextID[k] := 'a';
        ELSIF NextID[k] = '9' THEN
            NextID[k] := 'A';
        ELSE
            INC (NextID[k]);
        END (*IF*);
    END IncrNextID;

(************************************************************************)

PROCEDURE GenerateSessionID (VAR (*OUT*) sessID: sessIDtype);

    (* Assigns a session ID to a new session. *)

    BEGIN
        Obtain (NextIDLock);
        sessID := NextID;
        IncrNextID (lastAN);
        Release (NextIDLock);
    END GenerateSessionID;

(************************************************************************)

PROCEDURE AppendSessID (VAR (*IN*) ID: ARRAY OF CHAR;
                                    VAR (*INOUT*) message: ARRAY OF CHAR);

    (* Appends ID, as an ASCII string, to message. *)

    BEGIN
        Strings.Append (ID, message);
    END AppendSessID;

(************************************************************************)
(*                     SAVING THE CURRENT SESSION ID                    *)
(************************************************************************)

PROCEDURE SaveSessionParams (VAR (*IN*) sessID: ARRAY OF CHAR;
                                  ciphersuite: CipherSuite;
                                    VAR (*IN*) mastersecret: ARRAY OF CARD8);

    (* Saves a copy of the current data for sessID.    *)

    VAR p, q: SavedSess;  found: BOOLEAN;

    BEGIN
        Obtain (sesslist.access);
        p := FindID (sessID);
        found := p <> NIL;
        IF found THEN
            (* We'll overwrite the old values. *)
        ELSE
            (* Delete last list element if the list is too long. *)

            IF sesslist.count >= maxcount THEN
                DEC (sesslist.count);
                q := sesslist.tail;
                IF q <> NIL THEN
                    q^.prev^.next := NIL;
                    sesslist.tail := q^.prev;
                    DISPOSE (q);
                END (*IF*);
            END (*IF*);

            (* Now put the new entry at the head of the list. *)

            INC (sesslist.count);
            NEW (p);
            Strings.Assign (sessID, p^.sessID);
            p^.prev := NIL;
            p^.next := sesslist.head;
            sesslist.head := p;
            IF p^.next = NIL THEN
                sesslist.tail := p;
            END (*IF*);
        END (*IF*);
        p^.suite := ciphersuite;
        Copy (ADR(mastersecret), ADR(p^.master), mslength);
        Release (sesslist.access);
    END SaveSessionParams;

(************************************************************************)
(*                   RESTORING SAVED ID INFORMATION                     *)
(************************************************************************)

PROCEDURE CheckSessionID (VAR (*IN*) sessID: ARRAY OF CHAR;
                           VAR (*OUT*) ciphersuite: CipherSuite;
                           VAR (*OUT*) mastersecret: ARRAY OF CARD8): BOOLEAN;

    (* Returns FALSE if a matching seeionID cannot be found.  If it is  *)
    (* found, returns TRUE with the saved ciphersuit and mastersecret.  *)

    VAR found: BOOLEAN;  p: SavedSess;

    BEGIN
        Obtain (sesslist.access);
        p := FindID (sessID);
        found := p <> NIL;
        IF found THEN
            ciphersuite := p^.suite;
            Copy (ADR(p^.master), ADR(mastersecret), mslength);
        END (*IF*);
        Release (sesslist.access);
        RETURN found;
    END CheckSessionID;

(************************************************************************)
(*                        MODULE INITIALISATION                         *)
(************************************************************************)

PROCEDURE SetINIname (name: ARRAY OF CHAR);

    (* Takes note of our INI name. *)

    VAR hini: INIData.HINI;
        SYSapp: ARRAY [0..7] OF CHAR;

    BEGIN
        Strings.Assign (name, ININame);
        hini := INIData.OpenINIFile(ININame);
        IF INIData.INIValid (hini) THEN
            SYSapp := '$SYS';
            IF NOT INIData.INIGetString (hini, SYSapp, 'NextID', NextID) THEN
                ResetNextID;
            END (*IF*);
            INIData.CloseINIFile (hini);
        END (*IF*);
    END SetINIname;

(************************************************************************)

PROCEDURE SaveNextID;

    (* Copies NextID to the INI file. *)

    VAR hini: INIData.HINI;
        SYSapp: ARRAY [0..7] OF CHAR;

    BEGIN
        hini := INIData.OpenINIFile(ININame);
        IF INIData.INIValid (hini) THEN
            SYSapp := '$SYS';
            INIData.INIPutString (hini, SYSapp, 'NextID', NextID);
            INIData.CloseINIFile (hini);
        END (*IF*);
    END SaveNextID;

(************************************************************************)

BEGIN
    CreateLock (NextIDLock);
    ResetNextID;
    CreateLock (sesslist.access);
    sesslist.head := NIL;
    sesslist.tail := NIL;
    sesslist.count := 0;
FINALLY
    SaveNextID;
END TLSsessID.

